#!/usr/bin/python3
# -*- coding: utf-8 -*-
import argparse
import os
import sys
import re
import time
import docx
#import gen_pdf
#import pdb
from KAPILib import ShellLog, NormalDecl,VarDeclaration, FuncDeclaration, Matcher
from collections import defaultdict
from pycparser import c_parser, c_ast, parse_file

timestamp = int(time.time())

parser = argparse.ArgumentParser(description="** 内核API兼容性检查工具 ** ：API规范文件在kapi目录下，使用时需要输入内核头文件根目录", prog="KAPIChecker")

# The kapi standard file, searching from './kapi'
parser.add_argument('-k', '--kapi', action='store', type=str, help='内核API规范文件名称(默认:TCESA-20241115)', default="TCESA-20241115.docx")

#在非loongarch架构下扫描loongarch架构接口需要开启该选项
parser.add_argument( '--loongarch', action="store_true",help='扫描loongarch架构')

# The Target linux-header source directory
parser.add_argument('-d', '--hdrdir', action='store', type=str, help='待检测的内核头文件根目录(不能为空)', default="")

args = parser.parse_args()

file_time = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(timestamp))
arch_name = os.popen("uname -m").read().rstrip("\n")
check_result = defaultdict(list)

shelllog = ShellLog()

def check_kapi(h, funs, hdrdir):
    # Splice the according header file.
    header = hdrdir + "/" + h
    if not os.path.isfile(header):
        #错误类型1：文件不存在
        shelllog.error(f"Can't open {header}\n")
        return None

    err = []
    matcher = Matcher()
    matcher.readCode(header)
    for f in funs:
        #错误类型2：接口检查失败
        if (matcher.match(f) == False):
            err.append([h, '标准定义', f])
            err.append(['', '检测定义', matcher.get_source_string()])
            err.append(['', '失败原因', matcher.get_failed_info()])
    return err

def check_loongarch_kapi(kapis_loongarch, hdrdir, count, failure):
    matcher = Matcher()
    match = re.search(r'(.*?/)include(/)*\s*$',hdrdir)
    if not match:
        shelllog.error(f"Can't get archdir from {hdrdir}")
        archdir = hdrdir
    else:
        archdir = match.group(1)
    for h in kapis_loongarch:
        header = archdir + "/" + h.strip()
        flag_fail = False
        if not os.path.isfile(header):
            shelllog.error(f"Can't open {header}\n")
            count['文件缺失'] += 1
            failure.append(['loongarch', '失败原因', f"{h.strip()}件不存在"])
            continue;
        result = matcher.analyse_c_source(kapis_loongarch[h])
        matcher.readCode(header, flag_def=0)
        for item in result:
            src = matcher.analyse_c_source(matcher.code, item.name)
            if  not src or src[0] != item:
                count['文件接口不符'] += 1
                flag_fail = True
                err = []
                err.append(['loongarch', '头文件', h.strip()])
                err.append(['', '检测定义', item.name])
                err.append(['', '失败原因', src[0].get_failed_info()])
                failure.extend(err)
            shelllog.debug(item)
        if not flag_fail:
            count['文件检测通过'] += 1

# Return a dict: header files and functions whithin it
def read_kapi_spec(doc_path):
    doc = docx.Document(doc_path)

    kapis_loongarch = defaultdict(str)
    key_loongarch = ''
    text_loongarch = ''

    begin, end = 0, 0
    for index in range(len(doc.paragraphs)):
        if not begin and doc.paragraphs[index].text == "A.1 LoongArch":
            begin = index
        elif begin and re.search('eBPF相关接口', doc.paragraphs[index].text):
            end = index
            break

    for index in range(begin,end):
        para = doc.paragraphs[index]
        match = re.search(r'A.1.\d*\s*(arch/loongarch.*)',para.text)
        if match:
            if key_loongarch:
                kapis_loongarch[key_loongarch]= text_loongarch
                text_loongarch = ''
            key_loongarch = match.group(1)
        elif key_loongarch:
            text_loongarch += f"{para.text}\n"
    # 处理最后一个数据
    if key_loongarch and text_loongarch:
        kapis_loongarch[key_loongarch] = text_loongarch

    kapis = defaultdict(list)

    for table in doc.tables:
        for row in table.rows[1:]:
            row_data = [cell.text.strip() for cell in row.cells]
            if (len(row_data) == 3 or len(row_data) == 4):
                kapis[row_data[1]].append(row_data[2])
    return kapis, kapis_loongarch

def create_report_pdf(path, data):
    timestamp, args, count, failure = data
    table_failure = [["<font size=20>检查失败的头文件</font>"],["头文件", "接口"]]
    table_failure.extend(failure)

    machine_info = os.popen("uname  -r -s -m").read().rstrip("\n")
    pdf = gen_pdf.GenPdf(path)
    pdf.draw_title("内核API兼容性检查报告")
    pdf.draw_header("一、概述")
    text_time = time.strftime("%Y年%m月%d日 %H:%M:%S", time.localtime(timestamp))
    text_hdr = re.search(r".*/([^/]*)/include",args.hdrdir).group(1)
    pdf.draw_sentence(f"检测时间： {text_time}")
    pdf.draw_sentence(f"运行系统信息： {machine_info}")
    pdf.draw_sentence(f"头文件版本： {text_hdr}")
    pdf.draw_sentence(f"API规范版本： {args.kapi}")
    text=f'共检测{sum(count.values())}个文件'
    labels, sizes = [], []
    for key in count:
        text += f'，{count[key]}个{key}'
        if count[key]:
            labels.append(key)
            sizes.append(count[key])
    pdf.draw_paragraph(text)
    pdf.draw_pie(labels, sizes, title="图1、头文件检测情况")
    pdf.draw_header("二、详细数据")
    pdf.draw_table_special(table_failure)
    pdf.build_pdf()
    abspath = os.path.abspath(path)
    shelllog.info(f"检测报告已生成： {abspath}")

def create_report_txt(path, data):
    timestamp, args, count, failure = data
    table_failure = [["检查失败的头文件", "头文件", "接口"]]
    table_failure.extend(failure)

    machine_info = os.popen("uname -r -s -m").read().rstrip("\n")

    # 创建一个临时文件对象来写入内容
    with open(path, 'w') as txt_file:
        txt_file.write("内核API兼容性检查报告\n")
        txt_file.write("一、概述\n")

        text_time = time.strftime("%Y年%m月%d日 %H:%M:%S", time.localtime(timestamp))
        text_hdr = re.search(r".*/([^/]*)/include",args.hdrdir).group(1)
        txt_file.write(f"检测时间： {text_time}\n")
        txt_file.write(f"运行系统信息： {machine_info}\n")
        txt_file.write(f"头文件版本： {text_hdr}\n")
        txt_file.write(f"API规范版本： {args.kapi}\n")
        txt_file.write(f"共检测{sum(count.values())}个文件\n")

        labels, sizes = [], []
        for key in count:
            txt_file.write(f"{count[key]}个{key}\n")
            if count[key]:
                labels.append(key)
                sizes.append(count[key])
        if failure:
            txt_file.write("\n二、详细数据\n")
            for row in table_failure:
                txt_file.write("\t".join(row) + "\n")
                txt_file.write("\n")

    abspath = os.path.abspath(path)
    shelllog.info(f"检测报告已生成： {abspath}")

def main_loop(doc, hdrdir):
    kapis, kapis_loongarch = read_kapi_spec(doc)
    count = {'文件检测通过':0, '文件接口不符':0, '文件缺失':0}
    failure = []
#    report_pdf = "report-" + file_time + ".pdf"
    report_txt = "report-" + file_time + ".txt"

    if re.search('loongarch', arch_name, re.I) or args.loongarch:
        check_loongarch_kapi(kapis_loongarch, hdrdir, count, failure)

    for header in kapis:
        if not header:
            shelllog.debug("main_loop: header is None")
            continue
        err = check_kapi(header, kapis[header], hdrdir)
        if err != None:
            if len(err) == 0:
                count['文件检测通过'] += 1
            else:
                count['文件接口不符'] += 1
                failure.extend(err)
        else:
            count['文件缺失'] += 1
            failure.append([header, '失败原因', "文件不存在"])
    shelllog.info("报告生成中......")
#    create_report_pdf(report_pdf, [timestamp, args, count, failure])
    create_report_txt(report_txt, [timestamp, args, count, failure])

if __name__ == '__main__':
    doc_path = 'kapi/' + args.kapi
    if args.hdrdir == "":
        shelllog.info("请输入内核头文件目录，使用 '-h' 查看帮助")
        sys.exit(0)
    if not os.path.isdir(args.hdrdir):
        shelllog.info("输入内核头文件目录不存在，使用 '-h' 查看帮助")
        sys.exit(0)

    header_dir = args.hdrdir
    main_loop(doc_path,header_dir)
