#!/usr/bin/python3
#依赖pip install reportlab matplotlib  svglib
#reportlab用于绘制pdf， matplotlib用于绘制图表, svglib用于将matplotlib输出的svg图片转换为 reportlab graphics
#支持输出文字段落 标题 matplotlib的矢量图（svg格式） 以及表格
import re
import math
import pickle
from io import BytesIO
import matplotlib.pyplot as plt
from reportlab.lib import colors
from svglib.svglib import svg2rlg
from reportlab.pdfgen import canvas
from reportlab.pdfbase import pdfmetrics
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.pdfbase.cidfonts import UnicodeCIDFont
from reportlab.platypus import Paragraph, SimpleDocTemplate, Table, TableStyle, Spacer, Frame, PageTemplate

class GenPdf:
    stylesheet = getSampleStyleSheet()
    style = TableStyle([
    ('BACKGROUND', (0, 0), (-1, 0), colors.white),  # 设置表头背景色
    ('TEXTCOLOR', (0, 0), (-1, 0), colors.black),  # 设置表头文本颜色
    ('ALIGN', (0, 0), (-1, -1), 'CENTER'),  # 设置所有单元格水平居中对齐
    ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'), # 设置所有单元格垂直居中对齐
    ('BOTTOMPADDING', (0, 0), (-1, 0), 12),  # 设置表头下方填充
    ('BACKGROUND', (0, 1), (-1, -1), colors.white),  # 设置表格主体背景色
    ('GRID', (0, 0), (-1, -1), 1, colors.black),  # 设置表格网格线
    ('SPAN', (0, 0), (-1, 0)) #合并表头
    ])
    width, height = A4
    centered_style = ParagraphStyle(
        name='CenteredText',
        parent=stylesheet['BodyText'],
        alignment=1  # 居中对齐
    )
    @staticmethod
    def font_syle(text):
        if re.compile(r'[\u4e00-\u9fff]').search(text):
            return Paragraph(f'<font name="STSong-Light">{text}</font>', GenPdf.centered_style)
        else:
            return text

    @staticmethod
    def config_chinese():
        #设置中文字体
        #https://matplotlib.org/stable/users/explain/text/fonts.html#font-fallback
        plt.rcParams['font.family'] = 'Noto Sans CJK JP'
        #https://docs.reportlab.com/reportlab/userguide/ch3_fonts/#asian-font-support
        pdfmetrics.registerFont(UnicodeCIDFont('STSong-Light'))

    @staticmethod
    def wrap_text(text, font_name, font_size, max_width):
        wrapped_text = ""
        current_line = ""
        for word in text.split():
            test_line = f"{current_line} {word}".strip()
            line_width = stringWidth(test_line, font_name, font_size)

            if line_width <= max_width:
                current_line = test_line
            else:
                wrapped_text += current_line + "\n"
                current_line = word

        wrapped_text += current_line
        return wrapped_text

    def __init__(self, path="result.pdf"):
        self.path = path
        self.pdf = []
        self.doc = SimpleDocTemplate(path, pagesize=A4)
        self.config_chinese()
        self.left_margin = self.doc.leftMargin
        self.right_margin = self.doc.rightMargin
        self.max_width = GenPdf.width - self.left_margin - self.right_margin

    def calculate_column_widths(self, data, max_width,font_name='Helvetica', font_size=10):
        #ratios设置了各个表格单元的长度比例, 超出长度的单词之间会插入空格
        ratios = [[1],[0.2,0.8],[0.2, 0.15, 0.65]]
        widths = [math.ceil(max_width * ratio) for ratio in ratios[-1]]
        for row in range(len(data)):
            for col in range(len(data[row])):
                width = ratios[len(data[row]) - 1][col] * self.max_width - 10
                data[row][col] = GenPdf.wrap_text(str(data[row][col]), "Helvetica", 10, width)
        return widths

    def draw_pie(self, labels, sizes, title=""):
        svg_file = BytesIO()
        plt.pie(sizes, labels=labels, autopct='%1.1f%%')
        plt.title(title)
        plt.legend()
        plt.savefig(svg_file, format='SVG')
        svg_file.seek(0)
        drawing = svg2rlg(svg_file)
        scale = 0.75
        drawing.width *= scale
        drawing.height *= scale
        drawing.scale(scale, scale)
        self.pdf.append(drawing)
        self.pdf.append(Spacer(1, 12))
        return self.pdf

    def draw_table(self, data):
        widths = self.calculate_column_widths(data, self.max_width)
        data = [[self.font_syle(value) for value in row] for row in data]
        table = Table(data, colWidths=widths)
        table.setStyle(self.style)
        self.pdf.append(table)
        self.pdf.append(Spacer(1, 30))
        return self.pdf

    '''
    绘制特殊格式的表格，特征如下:
    |  xxxx    |
    ------------
    |x|  xxx   |
    -----------
    |x|x| xxx  |
      ----------
      |x| xxx  |
      ----------
      |x| xxx  |
    ------------
    分别为1列的表头，两列的小表头，三列的正文表格，其中正文表格的左侧每三行合并一次
    '''
    def draw_table_special(self, data):
        widths = self.calculate_column_widths(data, self.max_width)
        table_style =[
        ('BACKGROUND', (0, 0), (-1, 0), colors.white),  # 设置表头背景色
        ('TEXTCOLOR', (0, 0), (-1, 0), colors.black),  # 设置表头文本颜色
        ('ALIGN', (0, 0), (-1, -1), 'CENTER'),  # 设置所有单元格水平居中对齐
        ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'), # 设置所有单元格垂直居中对齐
        ('BOTTOMPADDING', (0, 0), (-1, 0), 12),  # 设置表头下方填充
        ('BACKGROUND', (0, 1), (-1, -1), colors.white),  # 设置表格主体背景色
        ('GRID', (0, 0), (-1, -1), 1, colors.black),  # 设置表格网格线
        ('SPAN', (0, 0), (-1, 0)), #合并表头
        ('SPAN', (1, 1), (2, 1)), #合并小表头
        ]

        for i in range(2, len(data), 1):
            if data[i][0] == '' and data[i-1][0]== '':
                table_style.append(('SPAN', (0, i-2), (0, i))) #合并左侧相邻的3个单元

        data = [[self.font_syle(value) for value in row] for row in data]
        table = Table(data, colWidths=widths)
        table.setStyle(TableStyle(table_style))
        self.pdf.append(table)
        self.pdf.append(Spacer(1, 30))
        return self.pdf

    def draw_sentence(self, strings):
        normalStyle = self.stylesheet['BodyText']
        normalStyle.fontName = 'STSong-Light'
        normalStyle.fontSize=12
        paragraph = Paragraph(strings, normalStyle)
        self.pdf.append(paragraph)
        return self.pdf

    def draw_paragraph(self, strings):
        self.draw_sentence(strings)
        self.pdf.append(Spacer(1, 12))
        return self.pdf

    def draw_title(self, strings):
        normalStyle = self.stylesheet['title']
        normalStyle.fontName = 'STSong-Light'
        normalStyle.fontSize=40
        title = Paragraph(strings, normalStyle)
        self.pdf.append(title)
        self.pdf.append(Spacer(1, 30))
        return self.pdf

    def draw_header(self, strings):
        normalStyle = self.stylesheet['Heading1']
        normalStyle.fontName = 'STSong-Light'
        normalStyle.fontSize=25
        title = Paragraph(strings, normalStyle)
        self.pdf.append(title)
        self.pdf.append(Spacer(1, 25))
        return self.pdf

    def build_pdf(self):
        self.doc.build(self.pdf)
