#! /usr/bin/bash
#依赖包coreutils
#检查rpm包时，依赖包rpm2cpio（debian）/rpm(fedora)
#检查dpkg包时，依赖包dpkg
#检查iso文件时，除上述包外，还可能依赖包squashfs-tools

#调试开关
#set -x
#当前检查过的数量
CHECKED_NUM=0

usage() {
	echo "可以检测iso、内核头文件包、头文件目录"
	echo "不添加任何参数或者仅指定-k会检测当前系统头文件目录"
	echo "Usage: "
	echo "	--kapi|-k:	[API规范文件名]"
	echo "	--dir|-d:	[待检测的头文件目录]"
	echo "	--iso|-i:	[待检查的操心系统iso] #检查iso过程时会申请root权限, 可能需要用户进行交互操作"
	echo "	--pkg|-p:	[待检查的头文件rpm/deb包]"
	echo "	--all|-a	#检查所有版本的内核头文件，需要同--iso|-i选项一起使用"
	echo "	--kver		[需要检测的内核版本] #检查特定版本的内核头文件，需要同--iso|-i选项一起使用"
	echo "	--loongarch	[扫描loongarch接口 #在非loongarch架构上检测loongarch头文件时，需要开启该选项]"
	echo ""
	echo "当检测iso时，如果没有添加 -a 或者 -kver 选项, 在检索到多个内核版本的头文件rpm/deb后，会出现交互界面，需要
用户指定检测版本；如果没有发现rpm/deb/squashfs文件，会出现交互界面，需要用户指定iso中的头文件位置"
	echo "运行依赖python3，需要安装如下依赖包: pip install python-docx  pycparser reportlab matplotlib  svglib"
}
debug() {
	echo -e "\e[1;35mdebug:\e[0m $@"
}
error() {
	echo -e "\e[1;31merror:\e[0m $@" >&2
}
die() {
	error "$@"
	exit 2
}
check_cmd() {
	local command=$1
	local info=$2
	if  ! type $command > /dev/null 2>&1 ;then
		die "命令 $command 不存在， 建议安装 $info"
	fi
}
kapi_check() {
	local dir=$1
	local script='./KAPIChecker.py'
	local dir_opt=''
	if [ ! -f $script ];then
		pushd /opt/os-kapi-checker
		dir_opt='1'
	fi
	if [ "$KAP_PATH" ];then
		$script $LOONGARCH -k $KAP_PATH -d  $dir
	else
		$script $LOONGARCH -d $dir
	fi
	[ "$dir" == '1' ] && popd
	CHECKED_NUM=$((CHECKED_NUM + 1))
}
check_os() {
	local os_dir=/usr/src/kernels/*/include
	[ -f /usr/bin/dpkg ] && os_dir=/usr/src/linux-headers-*/include
	for dir in `realpath $os_dir`
	do
		kapi_check $dir
	done
}
extract_pkg_squa() {
	local file_type=$1
	local file=$2
	local path=$3
	echo "解压 $file 至 $path 中"
	if [ "$file_type" =  "rpm" ];then
		check_cmd rpm2cpio "deb包: rpm2cpio | rpm包: rpm"
		check_cmd cpio "包: cpio"
		rpm2cpio $file | cpio -diD $path
	elif [ "$file_type" = "deb" ];then
		check_cmd dpkg "包： dpkg"
		dpkg -x  $file $path
	else
		check_cmd unsquashfs  "包：squashfs-tools"
		sudo unsquashfs  -d $path/tmp $file
		sudo chmod -R 777  $path/tmp
	fi
}
choose_one() {
	local choice=$1
	select var in $choice "all" "quit";do
		case $var in
			all)
				echo "$choice"
				break ;;
			quit)
				break ;;
			"")
				continue ;;
			*)
				echo "$var"
				break ;;
		esac
	done
}
#成功返回0， 文件类型不匹配返回1，其余错误立即退出，返回值>1 并输出日志
process_pkg_squa() {
	local file="$1" kevr=$2 index="" file_type=("rpm" "deb" "squa")
	local str=("RPM" "Debian binary package" "Squashfs filesystem")
	local suffix=("/usr/src/kernels/*$kver*/include" "/usr/src/linux-headers-*$kver*/include")

	[ -f "$file" ] || die "包 $file 不存在，请指定正确的文件路径"
	#检查文件类型
	for i in ${!file_type[@]};do
		[ "`file -b $file | grep -i "${str[$i]}"`" ] && index=$i && break
	done
	[ "$index" ] || return 1

	local path=`mktemp -d` tmp_path=""
	[ -d "$path" ] || die "mktemp 命令无法创建临时目录"
	#解压文件
	extract_pkg_squa ${file_type[$index]} $file $path
	#获取头文件路径, deb/rpm
	if [[ "$index" < "2" ]];then
		tmp_path=${path}${suffix[$index]}
	else
		#squashfs, 处理deb/rpm
		tmp_path="${path}/tmp${suffix[0]}"
		[ -f ${path}/tmp/bin/dpkg ] && tmp_path="${path}/tmp${suffix[1]}"
	fi

	local check_dir=`realpath $tmp_path`
	for var in $check_dir;do
		echo "对 $file $var 执行kapi检查"
		kapi_check  $var
	done
	echo "清理临时目录 $path"
	rm -r $path
	return 0
}
#成功返回0，错误立即退出，返回值>1 并输出日志
process_iso() {
	local iso=$1
	local path=`mktemp -d`
	[ -f "$iso" ] || die "iso $iso 不存在，请指定正确的文件路径"
	[ -d "$path" ] || die "mktemp 命令无法创建临时目录"
	sudo mount $iso $path
	local tmp=`find $path -type f -name "kernel-devel*$KVER*.rpm" -o -name "linux-header*$KVER*.deb"`
	local times=0
	if [[ `echo "$tmp" | wc -l` > 1 && $FLAG_CHECK_ALL != 1 ]];then
		echo "存在多个版本，请选择："
		tmp=`choose_one "$tmp"`
	fi
	for file in $tmp;do
		process_pkg_squa $file && times=$((times + 1))
	done

	if [[ $times == 0 ]];then
		#squashfs, 尝试镜像中尺寸最大的文件
		local squa=`find $path -type f  | xargs du | sort -g | tail -1 | cut  -f 2`
		process_pkg_squa $squa $KVER && times=$((times + 1))
	fi
	while [ $times = "0" ];do
		error "镜像文件 $iso 中没有发现内核头文件"
		[ "$pkg" ] && echo -e "已搜索的package列表:\n$pkg"
		[ "$squa" ] && echo "已搜索的squashfs列表:\n$squa"
		echo "内核头文件一般存在如下文件中"
		if [ "$KVER" ];then
			echo "kernel-devel*$KVER*.rpm， linux-header*$KVER*.deb, *squashfs*"
			echo "可能--kver $KVER 选项存在错误，可以尝试移除--kver"
		else
			echo "kernel-devel*.rpm， linux-header*.deb, *squashfs*"
		fi
		local file_path
		read -q "如果镜像 $iso 中存在尚未搜索过的内核头文件[deb/rpm/squashfs]，请指定位置, 退出请输入q:"  file_path
		[ "file_path" = "q" ] && break;
		process_pkg_squa $file_path  && break;
	done
	sudo umount $path
	rmdir $path
}

#参数解析
PARA_NUM=$#
if [[ "$PARA_NUM" == "0" ]];then
	check_os
	exit 0
fi
while true;do
	case "$1" in
		--kapi|-k)
			KAP_PATH="$2"
			shift ;;
		--dir|-d)
			HEADER_PATH+="$2"
			shift ;;
		--iso|-i)
			ISO_PATH+=" $2"
			shift ;;
		--pkg|-p)
			PKG_PATH+=" $2"
			shift ;;
		--all|-a)
			FLAG_CHECK_ALL=1 ;;
		--kver)
			KVER=$2
			shift ;;
		--loongarch)
			LOONGARCH='--loongarch' ;;
		*)
			[ "$1" ] && [ "$1" != "-h" ] && [ "$1" != "--help" ] &&\
			error "未知选项 $1， 请查看usage"
			usage
			exit 0;;
	esac
	shift
	[ "$1" ] || break
done
if [[ ($PARA_NUM == '2') && "$HEADER_PATH" ]];then
	check_os
	exit 0
fi
for var in $HEADER_PATH;do
	kapi_check $var
done
for var in $PKG_PATH;do
	process_pkg_squa $var || die "$var 文件类型错误，同rpm/deb类型不匹配"
done
for var in $ISO_PATH;do
	process_iso $var
done
